// cvGPU.cpp 
// A C style call dynamic library (DLL) to encapsulate GPU functions of OpenCV
// Copyright (C) 2012 by Giandomenico De Sanctis
// gidesay@yahoo.it

#include <vector>
#include <iostream>

#include "stdafx.h"
#include "cvGPU.h"
#include "opencv2/imgproc/imgproc_c.h"


using namespace cv;
using namespace cv::gpu;



typedef struct _GPUImage 
{
	cv::gpu::GpuMat* assMat;
	CvArr* assArr;
} GPUImage;


typedef std::vector<GPUImage*> memVect;
// GPU images cache
memVect pAllocatedImages;
int     nAllocatedImages=0;
BOOLEAN flgDownloadResult=TRUE;


inline void convertGpuMat2IplImg(const cv::gpu::GpuMat gMat , const  CvArr* iplImg)
{
	// do only if is active the download of GPU result to CvArr 
	if (flgDownloadResult) {
		// convert from GpuMat to Mat
		cv::Mat matdst = gMat;
		
		// convert from Mat to IplImage* 
		IplImage newipl = matdst;
		cvCopy(&newipl, (CvArr*) iplImg, NULL);
	};
	return;
}



cv::gpu::GpuMat* convertIplImg2GpuMat(const CvArr* image ) 
{
	GpuMat* ret = NULL;
	BOOLEAN flgFound = FALSE;

	// search the image in the GPU cache
	memVect::const_iterator costIter = pAllocatedImages.begin();
	for(; costIter != pAllocatedImages.end(); costIter++)
	{
		GPUImage* P = (GPUImage*)(*costIter);
		// if it's found, return the existing GpuMat, just pre-loaded with image data
		if (P->assArr == image)
		{
			flgFound = TRUE;
			ret = P->assMat;
			break;
		};
	}; 

	// else allocate a new GPU mat
	if (!(flgFound))
	{	
		GPUImage* p = (GPUImage*) malloc(sizeof(GPUImage)); 
		if (p) {
			// convert input from CvArr* to cv::Mat
			cv::Mat matobj = cv::cvarrToMat(image);
			// convert from Mat class to GpuMat 
			p->assMat = new GpuMat(matobj) ;
			// remember the image associated with GpuMat
			p->assArr = (CvArr*) image;

			pAllocatedImages.push_back(p);
			nAllocatedImages++;
			ret = p->assMat;
		};
	};
	return ret;
};

/* Activate or deactivate the download of GPU result to CvArr struct   */
CVDLL_API CDECL void cvGPUDownloadResult(BOOLEAN flg) 
{
	flgDownloadResult = flg;
}


CVDLL_API CDECL void cvGPURelease() 
{
    CV_FUNCNAME( "cvGPURelease" );

    __CV_BEGIN__;

	memVect::const_iterator costIter = pAllocatedImages.begin();
	for(; costIter != pAllocatedImages.end(); costIter++)
	{
		GPUImage* P = (GPUImage*)(*costIter);
		delete (P->assMat);
		free(P);
	};	
    
	pAllocatedImages.clear();

	nAllocatedImages = 0;

 
	__CV_END__;
};


C_EXTERN CVDLL_API int cvGPUGetCudaEnabledDeviceCount()
{
	return cv::gpu::getCudaEnabledDeviceCount();
}

/* Load an image into a GPU matrix    */
C_EXTERN CVDLL_API int cvGPUCreateImage( const CvArr* image )
{
	GpuMat* p = convertIplImg2GpuMat(image);
	if (p) {
		return nAllocatedImages;
	} else {
		return -1;
	};	
}

/* Download a GPU matrix into the associated CvArr struct   */
C_EXTERN CVDLL_API CDECL void cvGPUReadImage(const CvArr* image)
{
	// search the image in the GPU cache
	memVect::const_iterator costIter = pAllocatedImages.begin();
	for(; costIter != pAllocatedImages.end(); costIter++)
	{
		GPUImage* P = (GPUImage*)(*costIter);
		// if it's found, download the values to CvArr structure
		if (P->assArr == image)
		{
			convertGpuMat2IplImg(*P->assMat, image);	
			break;
		};
	};	
}



/* Release a GPU matrix    */
CVDLL_API CDECL void cvGPUReleaseImage(const CvArr* image)
{
	// search the image in the GPU cache
	memVect::const_iterator costIter = pAllocatedImages.begin();
	for(; costIter != pAllocatedImages.end(); costIter++)
	{
		GPUImage* P = (GPUImage*)(*costIter);
		// if it's found, delete the existing GpuMat
		if (P->assArr == image)
		{
			delete (P->assMat);
			free(P);
			pAllocatedImages.erase(costIter);
			break;
		};
	};	
}

static void convertConvKernel( const IplConvKernel* src, cv::Mat& dst, cv::Point& anchor )
{
    if(!src)
    {
        anchor = cv::Point(1,1);
        dst.release();
        return;
    }
    anchor = cv::Point(src->anchorX, src->anchorY);
    dst.create(src->nRows, src->nCols, CV_8U);

    int i, size = src->nRows*src->nCols;
    for( i = 0; i < size; i++ )
        dst.data[i] = (uchar)src->values[i];
}


/* Converts input array pixels from one color space to another */
CVDLL_API CDECL void  cvGPUCvtColor( const CvArr* srcarr, CvArr* dstarr, int code )
{
	CV_Assert( ((IplImage *)srcarr)->depth == ((IplImage *)dstarr)->depth );
  	
	cv::gpu::GpuMat* src = convertIplImg2GpuMat(srcarr);
	cv::gpu::GpuMat* dst = convertIplImg2GpuMat(dstarr); 
	
	cv::gpu::cvtColor(*src, *dst, code,  0, cv::gpu::Stream::Null());

	convertGpuMat2IplImg(*dst, dstarr);	
}


/* dilates input image (applies maximum filter) one or more times.
   If element pointer is NULL, 3x3 rectangular element is used */
C_EXTERN CVDLL_API void  cvGPUDilate( const CvArr* srcarr, CvArr* dstarr,
                       IplConvKernel* element,
                       int iterations  )
{
    CV_Assert( ((IplImage *)srcarr)->width == ((IplImage *)dstarr)->width 
			&& ((IplImage *)srcarr)->height == ((IplImage *)dstarr)->height 
			&& ((IplImage *)srcarr)->nChannels == ((IplImage *)dstarr)->nChannels );

    cv::gpu::GpuMat* src = convertIplImg2GpuMat(srcarr);
	cv::gpu::GpuMat* dst = convertIplImg2GpuMat(dstarr); 
	
	cv::Point anchor;
	cv::Mat kernel;
	convertConvKernel( element, kernel, anchor );

	cv::gpu::dilate(*src, *dst, kernel, anchor, iterations, cv::gpu::Stream::Null() );

	convertGpuMat2IplImg(*dst, dstarr);	
}

/* Performs complex morphological transformation */
C_EXTERN CVDLL_API void cvGPUMorphologyEx( const void* srcarr, void* dstarr, void*,
                IplConvKernel* element, int op, int iterations )
{
  	CV_Assert( ((IplImage *)srcarr)->width == ((IplImage *)dstarr)->width 
			&& ((IplImage *)srcarr)->height == ((IplImage *)dstarr)->height 
			&& ((IplImage *)srcarr)->nChannels == ((IplImage *)dstarr)->nChannels );

	cv::gpu::GpuMat* src = convertIplImg2GpuMat(srcarr);
    	cv::gpu::GpuMat* dst = convertIplImg2GpuMat(dstarr); 

    	cv::Point anchor;
	cv::Mat kernel;
	IplConvKernel* temp_element = NULL;
	if (!element)
    	{
			temp_element = cvCreateStructuringElementEx(3, 3, 1, 1, CV_SHAPE_RECT);
    	} else {
    		temp_element = element;
    	};
    	convertConvKernel( temp_element, kernel, anchor );
    	if (!element)
    	{
    		cvReleaseStructuringElement(&temp_element);
    	};
//  void morphologyEx( const GpuMat& src, GpuMat& dst, int op, const Mat& kernel, Point anchor = Point(-1, -1), int iterations = 1, Stream& stream = Stream::Null());
    	cv::gpu::morphologyEx( *src, *dst, op, kernel, anchor, iterations, cv::gpu::Stream::Null() );

    	convertGpuMat2IplImg(*dst, dstarr);	
}

/* Resizes image (input array is resized to fit the destination array) 
   Supports CV_INTER_NN, CV_INTER_LINEAR
   supports CV_8UC1, CV_8UC4 types */
C_EXTERN CVDLL_API void cvGPUResize( const CvArr* srcarr, CvArr* dstarr, int method )
{
  	CV_Assert( ( ((IplImage *)srcarr)->nChannels == 1) ||   ( ((IplImage *)srcarr)->nChannels == 4) );
  	CV_Assert( ( ((IplImage *)dstarr)->nChannels == 1) ||   ( ((IplImage *)dstarr)->nChannels == 4) );
  	CV_Assert( ( ((IplImage *)srcarr)->depth == 8)  );
  	CV_Assert( ( ((IplImage *)dstarr)->depth == 8)  );
	CV_Assert( ( method  == CV_INTER_LINEAR) || ( method  == CV_INTER_NN) );				
	
	cv::gpu::GpuMat* src = convertIplImg2GpuMat(srcarr);
    cv::gpu::GpuMat* dst = convertIplImg2GpuMat(dstarr); 

	Size dsize = Size(dst->cols, dst->rows);
 	cv::gpu::resize(*src, *dst, dsize, 0, 0, method, cv::gpu::Stream::Null());

	convertGpuMat2IplImg(*dst, dstarr);	
}

/* Applies fixed-level threshold to grayscale image */
C_EXTERN CVDLL_API double  cvGPUThreshold( const CvArr*  srcarr, CvArr*  dstarr,
                            double  threshold, double  max_value,
                            int threshold_type )
{
  	CV_Assert( ((IplImage *)srcarr)->width == ((IplImage *)dstarr)->width 
			&& ((IplImage *)srcarr)->height == ((IplImage *)dstarr)->height 
			&& ((IplImage *)srcarr)->nChannels == ((IplImage *)dstarr)->nChannels
			&& ((IplImage *)srcarr)->nChannels == 1);
  	CV_Assert(  (((IplImage *)srcarr)->depth == 8)  
			&& (((IplImage *)dstarr)->depth == 8) ); 

	cv::gpu::GpuMat* src = convertIplImg2GpuMat(srcarr);
	cv::gpu::GpuMat* dst = convertIplImg2GpuMat(dstarr); 
  	double ret = cv::gpu::threshold(*src, *dst, threshold, max_value, threshold_type,  cv::gpu::Stream::Null());

	convertGpuMat2IplImg(*dst, dstarr);	
	return ret;
}

/* Performs generic geometric transformation using the specified coordinate maps */
C_EXTERN CVDLL_API void  cvGPURemap( const CvArr* srcarr, CvArr* dstarr,
                      const CvArr* mapxarr, const CvArr* mapyarr,
                      int flags ,
                      void* fillval )
{
	cv::gpu::GpuMat* src = convertIplImg2GpuMat(srcarr);
	cv::gpu::GpuMat* dst = convertIplImg2GpuMat(dstarr); 
	cv::Mat  dst0 = *dst;

	//    cv::Mat mapx = cv::cvarrToMat(_mapx), mapy = cv::cvarrToMat(_mapy);
	cv::gpu::GpuMat* mapx = convertIplImg2GpuMat(mapxarr);
	cv::gpu::GpuMat* mapy = convertIplImg2GpuMat(mapyarr); 

	CV_Assert( src->type() == dst->type() && dst->size() == mapx->size() );
    	CV_Assert(flags == CV_INTER_LINEAR);
	CV_Assert( (src->type() == CV_8UC1) ||  (src->type() == CV_8UC3) );
	CV_Assert( (mapx->type() == CV_32FC1) &&  (mapy->type() == CV_32FC1) );

	//! DST[x,y] = SRC[xmap[x,y],ymap[x,y]] with bilinear interpolation.
        //! supports CV_8UC1, CV_8UC3 source types and CV_32FC1 map type
// void remap(const GpuMat& src, GpuMat& dst, const GpuMat& xmap, const GpuMat& ymap);
	cv::gpu::remap( *src, *dst, *mapx, *mapy); 
	CV_Assert( dst0.data == dst->data );

	convertGpuMat2IplImg(*dst, dstarr);	
}


/*
BOOLEAN cvBlobDContour(CvSeq** seqCont, const CvContourChainCode blobListElem, CvMemStorage * memstor)
{
	BOOLEAN res = FALSE;


    CV_FUNCNAME( "cvBlobDContour" );

    __BEGIN__;

    if (( !seqCont ) || ( !(&blobListElem) )) 
        CV_ERROR( CV_StsNullPtr, "" );

	*seqCont = NULL;


	{ // init block	with iterators
	CvContourPolygon* pPolyg = cvConvertChainCodesToPolygon(&blobListElem);

	if (pPolyg->size() == 0) { delete pPolyg; return res; };

	// simplify a bit the contour
	CvContourPolygon* pContour = cvSimplifyPolygon(pPolyg, 5.0);
	delete pPolyg;

	CvSeq* seqpoints = NULL;
	seqpoints = cvCreateSeq(CV_SEQ_CONTOUR, sizeof(CvSeq), sizeof(CvPoint), memstor);

	CvPoint*  pNewPoint = (CvPoint*) cvBlobDAlloc(sizeof(CvPoint)*pContour->size());
    if (( !pNewPoint ) )
        CV_ERROR( CV_StsNullPtr, "Not enough memory for contour points" );
	CvContourPolygon::const_iterator pcontIter = pContour->begin();
	for(; pcontIter != pContour->end(); pcontIter++)
	{
		pNewPoint->x = (*pcontIter).x;
		pNewPoint->y = (*pcontIter).y;
        void* newElem = (void*)  cvSeqPush((CvSeq*) seqpoints, (void*) pNewPoint);
		pNewPoint++;
	};

	*seqCont = seqpoints;

	delete pContour;
	}; // end block with iterators


	__END__;


	return res;
};



CVBLOBDLL_API CDECL unsigned int cvBlobDLabel 
     (IplImage const *img, IplImage *imgOut, CvMemStorage * memstor, CvSeq** sblobs,
	 int minArea, int flags, const IplImage* colorImg)
{
	int res;
	CvBlobs lblobs;

	
    CV_FUNCNAME( "cvBlobDLabel" );

    __BEGIN__;

    if (( !memstor ) || ( !img ) || ( !imgOut ))
        CV_ERROR( CV_StsNullPtr, "" );
	if ( !CV_IS_STORAGE( memstor ))
		CV_ERROR( CV_StsBadArg, "1st param not a  CvMemStorage" );
	if ( !CV_IS_IMAGE( img ))
		CV_ERROR( CV_StsBadArg, "2st param not an IplImage" );
	if ( !CV_IS_IMAGE( imgOut ))
		CV_ERROR( CV_StsBadArg, "3st param not an IplImage" );
	if ( (colorImg) &&  (!CV_IS_IMAGE( colorImg )) )
		CV_ERROR( CV_StsBadArg, "colorImg param not an IplImage" );

	res = cvLabel(img, imgOut, lblobs);
	if (minArea<10) 
			minArea = 10;
	cvFilterByArea(lblobs, minArea, (img->width*img->height));

	if (lblobs.size() ==0)
	{
		sblobs = NULL;
        cvReleaseBlobs(lblobs);
		return 0;
	};
	CvSeq* seqblobs = NULL;
	seqblobs = cvCreateSeq(CV_SEQ_ELTYPE_GENERIC, sizeof(CvSeq), sizeof(CvBlobObj), memstor);
	
	CvBlobObj* pNewBlobObj = (CvBlobObj*) cvBlobDAlloc(sizeof(CvBlobObj)*lblobs.size());
   if (( !pNewBlobObj ) )
        CV_ERROR( CV_StsNullPtr, "Not enough memory for blobs" );
   for (CvBlobs::const_iterator it=lblobs.begin(); it!=lblobs.end(); ++it)
   {
	   CvBlob curBlob = (CvBlob) *((*it).second);
	   cvCentralMoments(&curBlob, imgOut);
			
	   memcpy(pNewBlobObj, &curBlob, sizeof(CvBlobObj));
		pNewBlobObj->angle = cvAngle(&curBlob);
		// eccentricity = ( (u20 - u02)^2 - 4 u11^2 ) / (u20 + u02)^2 );
		pNewBlobObj->eccentricity = 
			( (pNewBlobObj->u20 - pNewBlobObj->u02)*(pNewBlobObj->u20 - pNewBlobObj->u02) - 4*pNewBlobObj->u11*pNewBlobObj->u11 )
				/ ( (pNewBlobObj->u20 + pNewBlobObj->u02)*(pNewBlobObj->u20 + pNewBlobObj->u02) );
		pNewBlobObj->meanColor = cvScalarAll(0);
		pNewBlobObj->contour = NULL;
		pNewBlobObj->internalContours = NULL;

		if (flags & CVBLOBDLL_GETCONTOUR)
		{
			cvBlobDContour(&pNewBlobObj->contour, curBlob.contour, memstor);
		};

		if ((flags & CVBLOBDLL_GETHOLESCONT) && (curBlob.internalContours.size()>0))
		{
			CvSeq* precCont = NULL;
			CvSeq* curCont = NULL;
			CvContoursChainCode::const_iterator pcontIter = curBlob.internalContours.begin();
			for(; pcontIter != curBlob.internalContours.end(); pcontIter++)
			{
				// process only holes with at less MIN_HOLE_POINTS points
				if ( (*(*pcontIter)).chainCode.size()>=MIN_HOLE_POINTS) {
					cvBlobDContour(&curCont, *(*pcontIter) , memstor);
					if (precCont) {
						precCont->h_next = curCont;
					} else { 
						pNewBlobObj->internalContours = curCont;
					};
					curCont->h_prev = precCont;
					precCont = curCont;
				};
			};
		};
		if ((flags & CVBLOBDLL_GETMEANCOLOR) && (colorImg))  {
			pNewBlobObj->meanColor = cvBlobMeanColor(&curBlob, imgOut, colorImg);
		};
		
		void* newElem = (void*)  cvSeqPush((CvSeq*) seqblobs, (void*) pNewBlobObj);

		pNewBlobObj++;
  }



   *sblobs = seqblobs;
   cvReleaseBlobs(lblobs);
	
	__END__;


	return res;
};
*/

